/* *******************************************************************
   Zwei(?) Strukturen für die gefundene IPs und Services

   List of MDNS Service types: http://www.dns-sd.org/ServiceTypes.html

   searching for devices is done in following steps
   - MDNS SD (service discovery): can gather hostname and activated services
   - PING:
   - HTTP GET: when we have found an device with PING we check that device for an HTTP Server on Port 80.
   ***************************************************************** */

#pragma once
#include <ESP8266Ping.h>               // V1.1.0 by Daniele Colanardi - manual download from https://github.com/dancol90/ESP8266Ping
constexpr size_t maxSize = -1;
uint8_t cntPingOK = 0;
uint8_t cntPingNOK = 0;
uint8_t lastPingOctet = 0;             // octet of last pinged IP

struct {
  String hostname;       // MDNS result
  bool pingable = false;
  uint32_t lastSeen;     // the ping result
  bool hasMdnsService = false;     // has the host reported any Service
  bool hasWebsiteService = false;  // has the host a website - reported as service
  bool hasWebsiteCheck = false;    // has the host a validated website at port 80
} myhosts[256];          // the index will be used as last octet for a /24 net (subnet 255.255.255.0)

struct {
  uint8_t octet;         // last octet of the IP this services belongs to
  uint16_t port;         // reported port where service is available
  String servicename;    // the reported service
  uint32_t lastSeen;     // timestamp of report
} myservices[100];       // a (limited) list of services

/*
    if a service was already seen - update service
    if it is a new service - store service to list
    return the written slot or maxSize if not possible
*/
size_t addService(uint8_t octet, uint16_t port, String servicename) {
  for (size_t i = 0; i < sizeof(myservices) / sizeof(myservices[0]); i++) {
    if (myservices[i].octet == octet && myservices[i].servicename == servicename && myservices[i].port == port) {
      Serial.print("update slot "); Serial.println(i);
      myservices[i].lastSeen = millis();
      return (i);
    }
  }

  for (size_t i = 0; i < sizeof(myservices) / sizeof(myservices[0]); i++) {
    if (myservices[i].lastSeen == 0) {     // tbd: "or to old"
      Serial.print("new slot "); Serial.println(i);
      myservices[i].octet = octet;
      myservices[i].port = port;
      myservices[i].servicename = servicename;
      myservices[i].lastSeen = millis();
      return (i);
    }
  }
  return maxSize; // not possible
}

// discover MDNS hostnames, services & ports and store in data structure
void discoverServiceAndStore(const char * servicename, const char * protocol) {
  Serial.printf("[MDNS] Browsing service _%s._%s.local. ... ", servicename, protocol);
  int n = MDNS.queryService(servicename, protocol);
  if (n == 0) {
    Serial.printf("no services found\n");
  } else {
    Serial.print(n);
    Serial.printf(" service(s) found\n");
    for (int i = 0; i < n; i++) {
      // Print details for each service found
      Serial.print("  ");
      Serial.print(i + 1);
      Serial.print(": ");
      Serial.print(MDNS.hostname(i));  // returns String object
      Serial.print(" (");
      Serial.print(MDNS.IP(i));        // returns IPAddress
      Serial.print(":");
      Serial.print(MDNS.port(i));      // returns uint16_t
      Serial.println(")");

      myhosts[MDNS.IP(i)[3]].hostname = MDNS.hostname(i);
      myhosts[MDNS.IP(i)[3]].hostname.remove(myhosts[MDNS.IP(i)[3]].hostname.length() - 6, 6); // remove ".local"
      myhosts[MDNS.IP(i)[3]].hasMdnsService = true;
      if (!strcmp(servicename, "http")) myhosts[MDNS.IP(i)[3]].hasWebsiteService = true;
      size_t result = addService(MDNS.IP(i)[3], MDNS.port(i), servicename);
      if (result == maxSize) Serial.println(F("E83: no emtpy slot"));
    }
    Serial.println();
  }
}

// each time this function is called one MDNS SD call will be triggered
// during runtime we call the services one by one because one service discovery lasts the default timeout of 1 second
void mdnsTrigger(bool doAllAtOnce = false) {
  struct {
    const char *name;         // serviceName[32]
    const char *protocol;     // protoName[32]
  } services[] {
    {"http", "tcp"},          // webserver
    {"arduino", "tcp"},       // ESP with OTA
    {"smb", "tcp"},           // e.g. a NAS
    {"printer", "tcp"},
    {"esp", "tcp"},           // Extensis Server Protocol (not really ESPs - but used in some MDNS Library Examples)
  };
  static size_t current = 0;
  if (doAllAtOnce) {
    for (auto &i : services) discoverServiceAndStore(i.name, i.protocol);
  } else {
    discoverServiceAndStore(services[current].name, services[current].protocol);
    current++;
    if (current >= sizeof(services) / sizeof(services[0])) current = 0;
  }
}

// ping
void pingRun() {
  IPAddress currentIp = WiFi.localIP();                  // your /24 network IP. We assume a subnet of 255.255.255.0
  const byte tries = 1;                                  // tries per ping
  const byte ipToPing = 1;                               // how many IPs should be pinged in one iteration
  static byte from = 0;                                  // first IP to scan in this iteration
  uint16_t to = from + ipToPing;                         // last IP to scan in this iteration
  if (to >= 255) to = 255;
  for (byte i = from ; i < to; i++) {
    currentIp[3] = i;
    lastPingOctet = i;
    Serial.print(F("[PING] IP "));
    Serial.print(currentIp);
    if (Ping.ping(currentIp, tries)) {
      Serial.println(F("\t Success"));
      myhosts[i].pingable = true;
      myhosts[i].lastSeen = millis();
      cntPingOK++;

      // let's check if we can find a webserver
      if (checkClient(i) > 0)                    // any HTTP Response Codes (even  errors) are fine
        myhosts[i].hasWebsiteCheck = true;
      else
        myhosts[i].hasWebsiteCheck = false;
    } else {
      Serial.println(F("\t timeout"));
      myhosts[i].pingable = false;
      cntPingNOK++;
    }
  }
  from = to % 255;                   // set next start IP
  if (from == 0) {
    cntPingOK = 0;
    cntPingNOK = 0;
  }
  Serial.print(F("heap=")); Serial.println(ESP.getFreeHeap());
}

// Output stored data on Serial interface
void debugData() {
  Serial.println(F("debug data"));
  for (size_t i = 0; i < sizeof(myhosts) / sizeof(myhosts[0]); i++) {
    if (myhosts[i].hostname != "" || myhosts[i].pingable) {
      Serial.print(i); Serial.print(" \t");
      Serial.print(myhosts[i].hostname); Serial.print(" \t");
      Serial.print(myhosts[i].lastSeen);
      for (size_t j = 0; j < sizeof(myservices) / sizeof(myservices[0]); j++) {
        if (i == myservices[j].octet) {
          Serial.print(" \t");
          Serial.print(myservices[j].servicename);
          Serial.print(":");
          Serial.print(myservices[j].port);
        }
      }
      Serial.println();
    }
  }
}

// call in Setup()
void mdnsSetup(char * myhostname) {
  if (!MDNS.begin(myhostname)) {
    Serial.println("Error setting up MDNS responder!");
  }
  else {
    Serial.println("mDNS responder started");
    MDNS.addService("http", "tcp", 80); // Announce service - we have an webserver running
  }

  // init the internal storage with own data
  // the MDNS SD will not work on it's own IP, so we do it manually
  byte octet = WiFi.localIP()[3];
  myhosts[octet].hostname = myhostname;     // myhostname is local in
  myhosts[octet].pingable = true;           // let's assume ping would work
  myhosts[octet].hasMdnsService = true;
  myhosts[octet].hasWebsiteService = true;
  myhosts[octet].hasWebsiteCheck = true;
  addService(octet, 80, "http");            // I know for sure I have implemented a server on port 80 ;-)

  // this principle could be used to manually name clients in your network which don't provide a MDNS SD hostname
  // the server will show them only when they report MDNS or were pinged successfully
  myhosts[30].hostname = "PC"; // just as an example

  // on startup we check for all services
  mdnsTrigger(true);
}

// trigger several tasks needed to scan the network
void netscanTimer() {
  constexpr uint32_t interval = 60 * 1000UL;
  static uint32_t previousMillis = 0;
  if (millis() - previousMillis > interval) {  // check if MDNS is due
    mdnsTrigger();
    previousMillis = millis();
  }

  constexpr uint32_t intervalP = 3 * 1000UL;
  static uint32_t previousMillisP = 0;
  if (millis() - previousMillisP > intervalP) {  // check if PING is d
    pingRun();  // ping can last > 1000ms - there is a new pull request for an async ping -
    previousMillisP = millis();
  }
}
//
