/* *******************************************************************
   Webserver

   How it all works together:

   page   \             includes JavaScript j.js and f.css
          9.htm         just a debug page
   script j.js          the JavaScript requests the JSON
          json          returns data as JSON
          cmd           a resource doing actions and not returning content (just the http header 204 no content)
   css    f.css         a css stylesheet for all on flash (program) memory
   ***************************************************************** */

#pragma once
#include <ESP8266WebServer.h>          // for the webserver
#include "mdns.h"                      // will need the structures with MDNS and PING data
// will access variables from your configuration file

ESP8266WebServer server(80);           // an instance for the webserver

constexpr byte fetchInterval = 5;      // in seconds
constexpr uint16_t messageSize = 5000; // max size of server message in bytes

void handleNotFound() {
  // Output a "404 not found" page. It includes the parameters which comes handy for test purposes.
  Serial.println(F("D026 handleNotFound()"));
  String message;
  message.reserve(messageSize);
  message += F("404 - File Not Found\n"
               "URI: ");
  message += server.uri();
  message += F("\nMethod: ");
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += F("\nArguments: ");
  message += server.args();
  message += F("\n");
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}

void handle204() {
  server.send(204);                // this page doesn't send back content
}

// the top of the page till tag <main>
void addTop(String &message) {
  message =  F("<!DOCTYPE html>\n"
               "<html lang='en'>\n"
               "<head>\n"
               "<title>My LAN Devices</title>\n"
               "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\">\n"
               "<meta name=\"viewport\" content=\"width=device-width\">\n"
               "<link rel='stylesheet' type='text/css' href='f.css'>\n"
               "<script src='j.js'></script>\n"
               "</head>\n");
  message += F("<body>\n");
  message += F("<header>\n<h1>My LAN Devices</h1>\n"
               "<nav><p><a href=\"/\">[Home]</a> </p></nav>\n</header>\n"
               "<main>\n");
}

// the bottom of the page from tag </main> downwards
void addBottom(String &message) {
  uint32_t ss = millis() / 1000;
  message += F("</main>\n"
               "<footer>\n<p>");                 // The footer will display the uptime, the version of the sketch and the compile date/time
  if (ss > 604800) {
    message += F("<span id='week'>");
    message +=  ((ss / 604800) % 52);
    message += F("</span> weeks ");
  }
  if (ss > 86400) {
    message += F("<span id='day'>");
    message += ((ss / 86400) % 7);
    message += F("</span> days ");
  }
  if (ss > 3600) {
    message += F("<span id='hour'>");
    message += ((ss / 3600) % 24);
    message += F("</span> hours ");
  }
  message += F("<span id='min'>");
  message += ((ss / 60) % 60);
  message += F("</span> minutes ");
  message += F("<span id='sec'>");
  message += (ss % 60);
  message += F("</span> seconds since startup | Version " VERSION );
  message += F(" | <span style='white-space:nowrap'>" __DATE__ " " __TIME__ "</span>"
               " <br><span id='msg'></span>"              // just an empty space for some debug information, hearbeat etc
               "</p>\n</footer>\n</body>\n</html>");
  server.send(200, "text/html", message);
}

// add a span and a fixtext
void addSpan(String &message, String css, String text) {
  message += F("<span class='");
  message += css;
  message += F("'>");
  message += text;
  message += F("</span>");
}


// the html output
// finally check your output if it is valid html: https://validator.w3.org
// *** HOME ***  0.htm
void handlePage() {
  String message;
  message.reserve(messageSize);
  addTop(message);

  uint16_t cntHosts = 0, cntServices = 0;

  for (size_t i = 0; i < sizeof(myhosts) / sizeof(myhosts[0]); i++) {
    if (myhosts[i].hasMdnsService || myhosts[i].pingable) {
      cntHosts++;
      String thisIP = (String)WiFi.localIP()[0] + '.' + (String)WiFi.localIP()[1] + '.' + (String)WiFi.localIP()[2] + '.' + i;
      if (myhosts[i].hasWebsiteService || myhosts[i].hasWebsiteCheck) {
        message += F("<article onclick=\"javascript:location.href='http://");
        message += thisIP;
        message += F("'\">\n");
      }
      else
        message += F("<article>");
      message += F("<h2>");
      message += thisIP;
      message += F("</h2>\n");
      if (myhosts[i].hostname != "") {
        message += F("<h3>");
        message += myhosts[i].hostname;
        message += F("</h3>\n");
      }
      message += F("<p>");

      //message += (String)myhosts[i].hasWebsiteService + "/" + (String)myhosts[i].hasWebsiteCheck + " <br>"; // debug only

      for (size_t j = 0; j < sizeof(myservices) / sizeof(myservices[0]); j++) {
        if (i == myservices[j].octet) {
          cntServices++;
          message += myservices[j].servicename; 
          message += ":";
          message += myservices[j].port;
          message += F("<br>\n");
        }
      }
      message += F("</p>\n");
      message += F("<p class='icon'>");
      if (i == WiFi.gatewayIP()[3]) addSpan(message, "g", "G");   // Gateway
      if (i == WiFi.localIP()[3]) addSpan(message, "s", "S");     // Scanning Device
      if (myhosts[i].hasWebsiteService || myhosts[i].hasWebsiteCheck) addSpan(message, "w", "W");               // Website
      if (myhosts[i].hasMdnsService) addSpan(message, "b", "B");  // Bonjour available
      if (myhosts[i].pingable) addSpan(message, "p", "P");        // Pingable
      message += F("</p>\n");
      message += F("</article>\n");
    }
  }

  message += F("<article>\n"
               "<h2>Statistics</h2>"
               "<p>Hosts: ");
  message += cntHosts;
  message += F("<br>\n");
  message += F("Services: ");
  message += cntServices;
  message += F("<br>\n");
  message += F("Ping OK: ");
  message += cntPingOK;
  message += F("<br>\n");
  //message += F("Ping NOK: ");
  //message += cntPingNOK;
  //message += F("<br>\n");
  message += F("last IP: ");
  message += lastPingOctet;
  message += F("<br>\n");
  message += F("</article>\n");

  message += F("<article>\n"
               "<h2>Legend</h2>\n"
               "<p>\n"
               "<span class='g'>G</span> Gateway<br>\n"
               "<span class='s'>S</span> Scanning Device<br>\n"
               "<span class='w'>W</span> Website<br>\n"
               //"<span class='u'>U</span> UPnP<br>\n"  // @todo: not implemented yet
               //"<span class='i6'>6</span> IP6<br>\n"  // @todo: not implemented yet
               "<span class='b'>B</span> Bonjour<br>\n"
               "<span class='p'>P</span> Pingable<br>\n"
               "</p>\n"
               "</article>\n");

  addBottom(message);
  server.send(200, "text/html", message);
}

// *** Debug ***  9.htm
// just some value output
void handlePage9() {
  String message;
  message.reserve(messageSize);
  addTop(message);

  // add your content hier

  addBottom(message);
  server.send(200, "text/html", message);
}

void handleCss() {
  // output of stylesheet
  // this is a straight forward example how to generat a static page from program memory
  String message;
  message = F(":root{\n"                          // some values (e.g. colors) might be used several times in several classes, therefore as variables
              "--no-col-main: #1279b1;\n"         // color main theme, e.g. h1 background
              "--no-col-on: #28A745;\n"           // color for on, active
              "--no-col-accent: orange;\n"
              "--no-col-warning: #ffc108;\n"      // color for warnings "yellow"
              "--no-col-danger: #DC3445;\n"       // color for danger, errors, "red"
              "--no-col-bg: white;\n"             // color main background
              "--no-col-bg0: #00487A;\n"          // color accents from light to darker
              "--no-col-bg1: #D0D0D0;\n"
              "--no-col-bg2: #F5F5F5;\n"
              "--no-col-bg5: silver;\n"           // e.g.: footer background
              "--no-col-bgm: #263238;\n"          // darkest background, e.g. for modal windows
              "--no-col-font: white;\n"           // contrast to main color, e.g. in h1
              "--no-col-font5: dimgrey;\n"        // e.g.: footer text
              "}\n"
              "*{font-family:sans-serif}\n"
              "body{margin:10px;background:var(--no-col-bg)}\n"
              "p {margin:0;font-size:1em}\n"
              "h1, h2{color:white;background:var(--no-col-main);text-align:center}\n"
              "h1{margin:1px;padding:5px;font-size:1.2em}\n"
              "h2{margin:0 0 2px 0;font-size:1.1em}\n"
              "h3{margin:0 0 2px 0;font-size:1.0em}\n"
              "a{text-decoration:none;color:dimgray;text-align:center}\n"
              "main{text-align:center}\n"
              "article{vertical-align:top;display:inline-block;position:relative;margin:0.1em;padding:0.1em;border-style:solid;border-width:1px;border-color:var(--no-col-bg1);background-color:var(--no-col-bg2);width:9.5em;height:8em;text-align:left}\n"
              "article h2{margin-top:0;padding-bottom:1px}\n"
              "section{margin-bottom:0.2em;clear:both}\n"
              "table{border-collapse:separate;border-spacing:0 0.2em}\n"
              "th, td{background-color:var(--no-col-bg1)}\n"
              "footer p{margin-bottom:5px;color:var(--no-col-font5);background:var(--no-col-bg5);font-size:0.7em;text-align:center}\n"
              "nav{background-color:var(--no-col-bg5);margin:1px;padding:5px;font-size:0.8em}\n"
              "nav a{color:var(--no-col-font5);padding:10px;text-decoration:none}\n"
              "nav a:hover{text-decoration:underline}\n"
              "nav p{margin:0px;padding:0px}\n"
              "button, input[type='button'], input[type='submit'], .btn, .on, .off{margin-top:0;margin-bottom:0.2em;margin-left:4em;font-size:1.4em;background:var(--no-col-main);border-style:solid;border-width:1px;border-radius:.3rem;border-style:outset;width:5em;height:1.5em;color:var(--no-col-font);text-decoration:none;text-align:center}\n"
              ".size2{display:flex;width:7em;height:1.5em;justify-content:center;align-items:center}\n"
              ".on{background:var(--no-col-on)}\n"
              ".off{background:var(--no-col-danger)}\n"
              "p.icon{position:absolute;top:6.9em}\n"
              ".g, .s, .w, .b, .u, .i6, .p {display:inline-block;margin:0 2px 1px 2px;text-align:center;color:white;background:grey;width:1em;border-radius:3px}\n"
              ".g{color:white;background:#FE5722}\n"/* gateway red*/
              ".s{color:white;background:#FFB200}\n"/* scanning orange*/
              ".w{color:white;background:#00BCD5}\n"/* website cyan*/
              ".b{color:white;background:#795547}\n"/* bonjour brown*/
              //".u{color:white;background:#3F51B5}\n"/* upnp blue*/
              //".i6{color:white;background:#AA47BC}\n" /* ip6 violette*/
              ".p{color:white;background:#4CB050}\n"/* pingable Grün*/
             );
  server.send(200, "text/css", message);
}

// Output: send data to browser as JSON
// after amodification always check if JSON is still valid. Just call the JSON (json) in your webbrowser and check.
// Serial.println(F("D266 requested json"));
void handleJson() {
  String message = "";
  message = (F(" {\"ss\":"));          // Start of JSON and the first object "ss":
  message += millis() / 1000;
#ifdef ARDUINO_ARCH_ESP8266            // The ESP8266 can meassure the internal voltage
  message += (F(",\"internalVcc\":"));
  message += ESP.getVcc();
#endif

  message += (F("}"));                           // End of JSON
  server.send(200, "application/json", message); // set MIME type https://www.ietf.org/rfc/rfc4627.txt
}


void handleCommand() {
  // receive command and handle accordingly
  // Serial.println(F("D283 handleCommand"));
  for (uint8_t i = 0; i < server.args(); i++) {
    Serial.print(server.argName(i));
    Serial.print(F(": "));
    Serial.println(server.arg(i));
  }
  if (false) ;                                        // needed just in case the board would not have any output pins


  else if (server.argName(0) == "CMD" && server.arg(0) == "RESET")   // Example how to reset the module. Just send ?CMD=RESET
  {
    Serial.println(F("D294 will reset"));
    ESP.restart();
  }

  server.send(204);                      // this page doesn't send back content --> 204
}


void handleJs() {
  // Output: a fetch API / JavaScript
  // a function in the JavaScript uses fetch API to request a JSON file from the webserver and updates the values on the page if the object names and ID are the same
  String message;
  message += F("const url ='json';\n"
               "function renew(){\n"
               " document.getElementById('sec').style.color = 'blue';\n"                                            // if the timer starts the request, the second gets blue
               " if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&orarr;';\n"         // show activity: "retrive request"
               " fetch(url)\n" // Call the fetch function passing the url of the API as a parameter
               " .then(response => {return response.json();})\n"
               " .then(jo => {\n"
               "   document.getElementById('sec').innerHTML = Math.floor(jo['ss'] % 60);\n"                         // example how to change a value in the HTML page
               "   for (var i in jo)\n"
               "   {\n"
               "    if (document.getElementById(i)) document.getElementById(i).innerHTML = jo[i];\n"                // as long as the JSON name fits to the HTML id, the value will be replaced
               "    if (document.getElementById('value'+i)) document.getElementById('value'+i).value = pwm2range(jo[i]);\n"    // in some cases we force to update the value, here for the PWM slider
               "    if (document.getElementById('cb'+i)) if(jo[i] == 1) document.getElementById('cb'+i).checked = true; else document.getElementById('cb'+i).checked = false;\n"
               "   }\n"  // update checkboxe
               // add other fields here (e.g. the delivered JSON name doesn't fit to the html id
               // finally, update the runtime
               "   if (jo['ss'] > 60) {document.getElementById('min').innerHTML = Math.floor(jo['ss'] / 60 % 60);}\n"
               "   if (jo['ss'] > 3600) {document.getElementById('hour').innerHTML = Math.floor(jo['ss'] / 3600 % 24);}\n"
               "   if (jo['ss'] > 86400) {document.getElementById('day').innerHTML = Math.floor(jo['ss'] / 86400 % 7);}\n"
               "   if (jo['ss'] > 604800) {document.getElementById('week').innerHTML = Math.floor(jo['ss'] / 604800 % 52);}\n"
               //"   document.getElementById('sec').style.color = 'dimgray';\n"                                     // if everything was ok, the second will be grey again.
               "   document.getElementById('sec').style.color = '';\n"                                              // if everything was ok, the second will be default color again
               "   if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&#10003;';\n"      // show activity: "OK"
               " })\n"
               " .catch(function() {\n"                                        // this is where you run code if the server returns any errors
               "  document.getElementById('sec').style.color = 'red';\n"
               " if (document.getElementById('msg')) document.getElementById('msg').innerHTML='&#10005;';\n"        // show activity: "offline"
               " });\n"
               "}\n"
               "document.addEventListener('DOMContentLoaded', renew, setInterval(renew, ");
  message += fetchInterval * 1000;
  message += F("));\n");

  message += F("function firecb(c){\n"                                 // c caller (mostly the id/self)
               " var request=new XMLHttpRequest();\n"
               " var v=0;\n"                                           // if value is initialisied with 0 we get ...
               " if (document.getElementById(c).checked) v=1;\n"       // ... 1 or 0 if checkbox is checked (better than true or false)
               " request.open('GET','cmd?cb='+c+'&v='+v,true);\n"
               " request.send(null);\n"
               "}\n");

  server.send(200, "text/javascript", message);
}

void serverSetup() {
  server.on("/",      handlePage);               // send root page
  server.on("/9.htm", handlePage9);              // output of values - can be used for debugging
  server.on("/f.css", handleCss);                // a stylesheet
  server.on("/j.js",  handleJs);                 // javascript based on fetch API to update the page
  server.on("/json",  handleJson);               // send data in JSON format
  server.on("/cmd",   handleCommand);            // process commands
  server.on("/favicon.ico", handle204);          // not available, send 204
  server.onNotFound(handleNotFound);             // show a typical HTTP Error 404 page
  server.begin();                                // start the webserver
}
//
